#include "stdafx.h"

#include "j2linfo.h"
#include <zlib.h>

/*
* This function takes a stream pointer and decompresses the data pointed by its
* compData field into a buffer which will be pointed to by the uncompData field.
* If it fails, st->uncompData and st->uncompSize will be in an unknown state.
* It's the caller's responsability to free the uncompressed data buffer by calling
* destroyStream() or destroyJ2FFile()
*/
int uncompressStream(cstreamP st)
{
	unsigned long	ucs = st->uncompSize;
	int				resuncomp; //to hold the result of uncompress

	if (st->compData == NULL) return STRM_NO_DATA; //error if there's no compressed data
	//if there's already uncompressed data, free it before
	if (st->uncompData != NULL) free(st->uncompData);
	st->uncompData = calloc(1, st->uncompSize);
	if (st->uncompData == NULL) return STRM_OUT_OF_MEMORY;

	resuncomp = uncompress(st->uncompData, &ucs, st->compData, st->compSize);

	if (resuncomp != Z_OK) {
		return STRM_OPERATION_FAILED; //uncompress failed
	}
	//the uncompressed data turned out be less than what was expected. Fail in this case
	if (ucs != st->uncompSize) return STRM_UNEXPECTED_SIZE;

	return STRM_SUCCESS;
}

/*
* Analogous to uncompressStream.
* If it fails, st->compData and st->compSize will be in an unknown state.
* It's the caller's responsability to free the compressed data buffer
* by calling destroyStream() or destroyJ2FFile().
*/
int compressStream(cstreamP st)
{
	if (st->uncompData == NULL)
		return STRM_NO_DATA; //error if there's no uncompressed data
	//if there's already compressed data, free it before
	if (st->compData != NULL)
		free(st->compData);
	//preliminary size, as per compress() documentation
	st->compSize = (int)ceil((double)st->uncompSize*1.001)+12;
	st->compData = calloc(1, st->compSize);
	if (st->compData == NULL)
		return STRM_OUT_OF_MEMORY;

	if (compress(st->compData, &st->compSize, st->uncompData, st->uncompSize)
		!= Z_OK) {
			return STRM_OPERATION_FAILED; //compress failed
	}

	return STRM_SUCCESS;
}

/*
* Constructors and destructors.
*/
cstreamP newStream(void)
{
	cstreamP		ret;

	ret = calloc(1, sizeof *ret);
	if (ret == NULL)
		return NULL;

	return ret;
}

void destroyStream(cstreamP stream)
{
	if (stream->compData != NULL)
		free(stream->compData);
	if (stream->uncompData != NULL)
		free(stream->uncompData);
	free(stream);
}

J2FFileP newJ2FFile(void)
{
	J2FFileP		ret;
	ret = calloc(1, sizeof *ret);	
	if (ret == NULL)
		return NULL;

	return ret;
}

void destroyJ2FFile(J2FFileP head)
{
	int i;
	for (i = 0; i <= 3; i++) {
		if (head->stream[i] != NULL)
			destroyStream(head->stream[i]);
	}
	free(head);
}

/* This is function is given a file pointer to a J2F file at offset 0
* and populates the prologue of the J2F file
*  See also readJ2FFile()
*/
static int readPrologueJ2FFile(J2FFileP head, FILE *f) {
	//initially only for levels, I shan't bother changing it
	if (fread(head->copyrightMessage,
		sizeof *(head->copyrightMessage) * 180, 1, f) == 0)
		return J2F_FAILURE;
	head->copyrightMessage[179] = '\0';
	if (fread(head->magic, sizeof *(head->magic) * 4, 1, f) == 0)
		return J2F_FAILURE;
	if (fread(head->passwordHash, sizeof *(head->passwordHash) * 3, 1, f) == 0)
		return J2F_FAILURE;
	if (fread(&head->hideLevel, sizeof (head->hideLevel), 1, f) == 0)
		return J2F_FAILURE;
	if (fread(head->levelName, sizeof *(head->levelName) * 32, 1, f) == 0)
		return J2F_FAILURE;
	head->levelName[31] = '\0';
	if (fread(&head->version, sizeof head->version, 1, f) == 0)
		return J2F_FAILURE;
	if (fread(&head->fileSize, sizeof head->fileSize, 1, f) == 0)
		return J2F_FAILURE;
	if (fread(&head->CRC32, sizeof head->CRC32, 1, f) == 0)
		return J2F_FAILURE;

	return J2F_SUCCESS;
}



/*
* Reads from a file to populate a J2FFile struct and decompress the streams.
* Last argument is J2F_TYPE_LEVEL or J2F_TYPE_TILESET.
* It's the caller's responsability to free what this function allocates
* (directly and through uncompressStream()) by calling destroyJ2FFile()
* That must be done even if this function fails.
*/
int readJ2FFile(const wchar_t *lvl, J2FFileP head, int type) {
	int			i;
	int			ret = -1;
	FILE		*file = NULL;

	if ((file = _wfopen(lvl, L"rb")) == NULL) {
		ret = J2F_COULD_NOT_OPEN_FILE;
		goto err;
	}

	//Populate J2FFile struct fields.
	//I will not fread directly into a struct as that is not portable (because
	//of data alignment).
	if (readPrologueJ2FFile(head, file) != J2F_SUCCESS) {
		ret = J2F_INVALID_FILE;
		goto err;
	}

	/* Make sure the header says it's a level or a tileset */
	if ((memcmp(head->magic, "LEVL", 4) != 0) && (type == J2F_TYPE_LEVEL)) {
		ret = J2F_INVALID_FILE;
		goto err;
	}
	if ((memcmp(head->magic, "TILE", 4) != 0) && (type == J2F_TYPE_TILESET)) {
		ret = J2F_INVALID_FILE;
		goto err;
	}

	for (i=0; i<=3; i++) {
		if (head->stream[i] != NULL) destroyStream(head->stream[i]);
		head->stream[i] = newStream(); //allocates the memory for the streams struct
		if (head->stream[i] == NULL) { //memory allocation failed
			ret = J2F_OUT_OF_MEMORY;
			goto err;
		}

		if (fread(&(head->stream[i]->compSize), sizeof(long), 1, file) == 0) {
			ret = J2F_INVALID_FILE;
			goto err;
		}
		if (fread(&(head->stream[i]->uncompSize), sizeof(long), 1, file) == 0) {
			ret = J2F_INVALID_FILE;
			goto err;
		}
	}

	for (i = 0; i <= 3; i++) { //uncompress the data
		//free compressed data if it already exists
		if (head->stream[i]->compData != NULL)
			free(head->stream[i]->compData);

		//allocate space for the compressed data
		head->stream[i]->compData = calloc(1, head->stream[i]->compSize);
		if (head->stream[i]->compData == NULL) {
			ret = J2F_OUT_OF_MEMORY;
			goto err;
		}

		//read compressed data
		if (fread(head->stream[i]->compData, head->stream[i]->compSize, 1, file)
			== 0) {
				//Could not read the compressed data of a stream.
				//Possible incorrect stream bounds or file truncated
				ret = J2F_INVALID_FILE;
				goto err;
		}

		//uncompressStream takes care of allocating space for the uncompressed data
		if (uncompressStream(head->stream[i]) != STRM_SUCCESS) {
			//Error uncompressing stream
			ret = J2F_INVALID_FILE;
			goto err;
		}
	}

	ret = J2F_SUCCESS;
err:
	if ((file != NULL) && (fclose(file) != 0)) {
		ret = J2F_FAILURE;
	}	

	return (ret == -1) ? J2F_FAILURE : ret;
}

#define STREAM_1_MIN_SIZE	(19+32*6)
int readLevelInformation(const wchar_t *file, LevelInfoP lvlinfo) {
	J2FFileP		head = NULL;
	int				temp;
	char			gendata[STREAM_1_MIN_SIZE];
	int				ret = -1,
					i;

	head = newJ2FFile();
	if (head == NULL) {
		ret = J2F_OUT_OF_MEMORY;
		goto err;
	}
	if ((temp = readJ2FFile(file, head, 1)) != 0) {
		ret = temp;
		goto err;
	}
	if (head->stream[0]->uncompSize < STREAM_1_MIN_SIZE) {
		ret = J2F_INVALID_FILE;
		goto err;
	}
	
	memcpy(gendata,head->stream[0]->uncompData, STREAM_1_MIN_SIZE);

	//make sure the strings are always null-terminated
	for (i = 1; i <= 6; i++) {
		//19 - skip initial irrelevant data (17 + 2, 2 being the version)
		//32 - the size of each of the relevant strings
		*(gendata+19+32*i-1) = '\x00';
	}
	lvlinfo->version = head->version;
	strcpy(lvlinfo->levelName,		gendata + 19);
	strcpy(lvlinfo->tileSet,		gendata + 19 + 32 * 1);
	strcpy(lvlinfo->bonusLevel,		gendata + 19 + 32 * 2);
	strcpy(lvlinfo->nextLevel,		gendata + 19 + 32 * 3);
	strcpy(lvlinfo->secretLevel,	gendata + 19 + 32 * 4);
	strcpy(lvlinfo->musicFile,		gendata + 19 + 32 * 5);

	ret = J2F_SUCCESS;
err:
	if (head != NULL)
		destroyJ2FFile(head);

	return (ret == -1) ? J2F_FAILURE : ret;
}

/*mIRC(LevelInformation) {
	J2FFileP		head;
	int				i;
	char			gendata[19+32*6],
	//the contents will be copied to data through mIRC_RETURN, and
	//data holds 1024 chars max, so no problem with 1024
					buf[1024];
	
	mIRC_PREAMBLE
	
	head = chkmal(newJ2FFile());
	if (head == NULL) goto err;
	if (readJ2FFile(data,head,1) != 0) {
		goto err;
	}

	if (head->stream[0]->uncompSize < 19+32*6) {
		ERR("Stream 1 is too short.");
		goto err;
	}
	else memcpy(gendata,head->stream[0]->uncompData,19+32*6);

	//make sure the strings are always null-terminated
	for (i = 1; i <= 6; i++) { *(gendata+19+32*i-1) = '\x00'; }

	sprintf(
		buf,
		"OK "
		"%i\x09"				//version
		"%s\x09"				//level name
		"%s\x09"				//tileset
		"%s\x09"				//bonus level
		"%s\x09"				//next level
		"%s\x09"				//secret level
		"%s",					//music file
		head->version,gendata+19,gendata+19+32*1,gendata+19+32*2,
		gendata+19+32*3,gendata+19+32*4,gendata+19+32*5
	);

err:
	if (head != NULL) destroyJ2FFile(head);
	
	mIRC_RETURN(buf)
}*/


